from .base_session import BaseSession


class PlainSession(BaseSession):

    BACKOFF_FACTOR = 1
    TOTAL_RETRIES = 3
    REQUEST_TIMEOUT = 4

    def make_request(self, url):
        super().make_request(url)
        return self.session.get(url, headers=self.headers)

    def make_stream_request(self, url):
        super().make_stream_request(url)
        return self.session.get(url, headers=self.headers, stream=True)



from fpdf import FPDF

import os
import shutil
from bikesanity.io_utils.resources import create_temp_from_resource


DEJAVU_FONT = 'DejaVu'


class JournalPdf(FPDF):

    A4_WIDTH = 210
    A4_HEIGHT = 297

    MARGIN = 20
    TOP_MARGIN = 10
    PAGE_WIDTH = A4_WIDTH - (MARGIN*2)

    IMAGE_SCALE = 1.2
    IMAGE_WIDTH = (A4_WIDTH - (MARGIN*2))/IMAGE_SCALE
    IMAGE_X = (PAGE_WIDTH/2) - (IMAGE_WIDTH/2)
    CAPTION_WIDTH = 150

    IMAGE_DPI = 200
    MM_PER_INCH = 25.4


    def __init__(self, title, author, part=None):
        self.draw_header = False
        self.journal_title = title
        self.author = author
        self.part = part

        super().__init__(orientation='P', unit='mm', format='A4')

        self.tmp_files = []
        self.load_font_resource('DejaVuSans.ttf', '')
        self.load_font_resource('DejaVuSans-Bold.ttf', 'B')
        self.load_font_resource('DejaVuSans-Oblique.ttf', 'I')
        self.page_title = title
        self.setup_pages()

    def load_font_resource(self, font_name, weight):
        # Get a temporary file from the named resource
        temp_font_file = create_temp_from_resource(['fonts', font_name])
        # Add the font from this temporary file (only method FPDF supports)
        self.add_font(DEJAVU_FONT, weight, temp_font_file, uni=True)
        # Remove the temp file once its loaded
        self.tmp_files.append(temp_font_file)


    def update_page_title(self, name):
        self.draw_header = False
        self.page_title = name

    def setup_pages(self):
        self.set_font(DEJAVU_FONT, '', 14)
        self.set_margins(self.MARGIN, self.TOP_MARGIN, self.MARGIN)
        self.add_page()

    def limit_title(self, title, max_width=PAGE_WIDTH):
        terms = title.split(' ')

        terms_to_use = []
        for i in range(0, len(terms)):
            terms_to_use.append(terms[i])
            title = ' '.join(terms_to_use)
            if self.get_string_width(title) > max_width: break

        return title

    def header(self):
        if not self.draw_header:
            self.draw_header = True
            return

        self.set_font(DEJAVU_FONT, 'B', 12)

        # Limit title if too long
        title = self.limit_title(self.page_title)

        # Calculate width of title and position
        w = self.get_string_width(title) + 6
        self.set_x((210 - w) / 2)

        # Title
        self.cell(w, 9, title, 0, 1, 'C', 0)
        # Line break
        self.ln(10)

    def footer(self):
        if self.page_no() < 3: return

        # Position at 1.5 cm from bottom
        self.set_y(-15)
        self.set_font(DEJAVU_FONT, 'I', 8)
        # Text color in gray
        self.set_text_color(128)

        footer_text = '{0} by {1}{2} - Page {3}'.format(self.journal_title, self.author, ' (part {0})'.format(self.part) if self.part else '', self.page_no()-2)

        # Page number
        self.cell(0, 10, footer_text, 0, 0, 'C')

    def cover_title(self, title, subtitle, author, distance_statement, part=None):
        self.set_font(DEJAVU_FONT, 'B', 20)

        self.ln(15)
        # Title
        self.multi_cell(0, 20, title, 0, 'C', 0)
        self.ln(1)
        if part:
            self.set_font(DEJAVU_FONT, '', 20)
            self.cell(0, 5, 'Part {0}'.format(part), 0, 0, 'C')
            self.ln(6)

        # Line break
        self.ln(6)
        self.set_font(DEJAVU_FONT, '', 16)
        self.multi_cell(0, 10, subtitle, 0, 'C', 0)
        self.ln(4)
        self.set_font(DEJAVU_FONT, 'I', 10)
        self.multi_cell(0, 5, distance_statement, 0, 'C', 0)
        self.ln(5)
        self.set_font(DEJAVU_FONT, '', 16)
        self.multi_cell(0, 20, author, 0, 'C', 0)
        self.ln(8)

    def section_title(self, title):
        self.set_font(DEJAVU_FONT, 'B', 18)
        self.multi_cell(0, 10, title, 0, 'C', 0)
        self.ln(20)

    def add_toc(self, toc_items):
        self.set_font(DEJAVU_FONT, 'B', 18)
        self.cell(0, 5, 'Table of Contents', 0, 0, 'C', 0)

        self.ln(15)
        self.set_font(DEJAVU_FONT, 'I', 9)

        for toc_item in toc_items:
            if toc_item.is_header:
                self.set_font(DEJAVU_FONT, 'B', 9)
            else:
                self.set_font(DEJAVU_FONT, 'I', 9)


            # Limit the title if it's too long
            title = self.limit_title(toc_item.title, 125)

            str_size = self.get_string_width(title)
            self.cell(str_size+2, 9, title)

            # Filling dots
            page_cell_size=self.get_string_width(toc_item.page_no) + 2

            dot_length = self.PAGE_WIDTH - str_size - page_cell_size - 10

            nb = int(dot_length // self.get_string_width('.'))
            dots = '.' * nb
            self.cell(dot_length, 9, dots, 0, 0, 'R')

            # Page number
            self.cell(page_cell_size, 9, toc_item.page_no, 0, 1,'R')

    def chapter_title(self, label, date, distance, total_distance):
        self.set_font(DEJAVU_FONT, 'B', 14)

        # Colors of frame, background and text
        self.set_draw_color(0, 0, 0)
        self.set_fill_color(230, 230, 0)
        # Thickness of frame (1 mm)
        self.set_line_width(0.5)

        # Background color
        self.set_fill_color(200, 220, 255)
        # Title
        self.multi_cell(0, 10, label, 1, 'C', 1)
        # Line break
        self.ln(4)

        if not distance: return

        if total_distance and date:
            distance_statement = '{0} - of total {1} - on {2}'.format(distance, total_distance, date)
        elif total_distance:
            distance_statement = '{0} - of total {1}'.format(distance, total_distance)
        else:
            distance_statement = distance

        self.set_font(DEJAVU_FONT, 'I', 10)
        self.cell(0, 5, distance_statement, 0, 0, 'L', 0)
        self.ln(20)

    def text_content(self, text):
        self.set_font("DejaVu", size=12)
        self.multi_cell(w=0, h=8, txt=text, align='J')
        self.ln(10)


    def try_add_image(self, image_path, height=None, ext=None):
        updated_ext = image_path[:image_path.rfind('.')] + ext if ext else image_path
        if ext and image_path != updated_ext:
            shutil.copyfile(image_path, updated_ext)
        try:
            if height:
                self.image(updated_ext, h=height)
            else:
                self.image(updated_ext, w=self.IMAGE_WIDTH)
            return True
        except:
            return False

    def add_image_format_tolerant(self, image_path, height=None):
        for ext in [ None, '.jpeg', '.png']:
            if self.try_add_image(image_path, height, ext):
                break

    def add_image(self, image_path, caption, height=None):
        self.cell(w=self.IMAGE_X, align='L')

        # Attempt to add the image in, trying different formats if the given extension is wrong (due to amateur treatment
        # in the source)
        self.add_image_format_tolerant(image_path, height)

        # Don't break between the image and the caption
        self.set_auto_page_break(False, margin=self.MARGIN)
        self.set_font("DejaVu", style='I', size=10)
        self.ln(5)
        self.multi_cell(w=self.PAGE_WIDTH, h=5, txt=caption, align='C')
        self.ln(9)
        self.set_auto_page_break(True, margin=self.MARGIN)

    def clipping_rect(self, x, y, w, h, outline=False):
        op= 'S' if outline else 'n'
        self._out('q {0} {1} {2} {3} re W {4}'.format(
            x * self.k,
            (self.h - y) * self.k,
            w * self.k,
            -h * self.k,
            op
        ))

    def unset_clipping(self):
        self._out('Q')

    def cleanup_tmp_files(self):
        for tmp_file in self.tmp_files:
            try:
                os.remove(tmp_file)
                os.remove(tmp_file + '.pkl')
                os.remove(tmp_file + '.cw127.pkl')
            except:
                pass



import os

import bikesanity.io_utils.log_handler as log_handler
from bikesanity.interpreter.journal_content import JournalContent
from bikesanity.interpreter.page_interpreter import PageInterpreter
from bikesanity.entities.journal import Journal
from bikesanity.entities.page import Page
from bikesanity.io_utils.local_journal import LocalJournalHandler
from bikesanity.io_utils.serializer import Serializer
from bikesanity.services.retrievers import DownloadingRetriever
from bikesanity.services.html_postprocessor import HtmlPostProcessor


class DownloadJournal:

    DOWNLOAD_DIRECTORY = 'downloads'

    def __init__(self, base_location, postprocess_html=True, progress_callback=None):
        self.base_location = os.path.join(base_location, self.DOWNLOAD_DIRECTORY)
        os.makedirs(self.base_location, exist_ok=True)

        self.serializer = Serializer()
        self.retriever = DownloadingRetriever()
        self.html_postprocessor = HtmlPostProcessor()

        self.journal_indexer = JournalContent(self.retriever)
        self.page_crawler = PageInterpreter(self.retriever, progress_callback=progress_callback)

        self.postprocess_html = postprocess_html
        self.progress_callback = progress_callback

    def progress_update(self, percent):
        if self.progress_callback:
            self.progress_callback(progress=percent)

    def get_download_location(self, journal) -> str:
        return os.path.join(self.base_location, journal.journal_id)

    def download_journal_url(self, url, from_page=0):
        try:
            # Retrieve the journal ID from the URL
            log_handler.log.info('Downloading journal {0}'.format(url))
            return self._download_journal_url(url, from_page)
        except Exception as exc:
            log_handler.log.exception('----> Error when processing journal {0}'.format(url))
            raise

    def _download_journal_url(self, url, from_page):

        log_handler.log.info('Resolving index page...')
        self.progress_update(percent=0)

        # Pull the title page and ToC
        journal = self.journal_indexer.retrieve_journal(url, None)
        if not journal: return None
        journal_id = journal.journal_id

        # Update progress
        self.progress_update(percent=5)
        log_handler.log.info('Successfully retrieved journal index. Processing journal ID {0}'.format(journal_id))

        local_journal_handler = LocalJournalHandler(base_path=self.base_location, journal_id=journal_id)

        # Retrieve the JS and CSS resources too
        journal = self.journal_indexer.retrieve_js_css_resources(journal)

        # Update progress
        self.progress_update(percent=10)
        log_handler.log.info('Successfully pulled JS and CSS resources')

        # Apply HTML post-processing for local browsability, if enabled
        if self.postprocess_html:
            self.html_postprocessor.postprocess_journal(journal)

        # Save original sources and clear resources
        journal.save_original_source(local_journal_handler)
        journal.clear_resources()

        # Download all the table of contents
        journal = self._process_journal(journal, local_journal_handler, from_page)

        self.progress_update(percent=100)

        return journal

    def _process_journal(self, journal: Journal, local_journal_handler, from_page):
        journal_id = journal.journal_id

        # Handle standard multi-page journals with a ToC
        if journal.toc:
            log_handler.log.info('Processing multiple pages for {0}'.format(journal_id))
            # Iterate over all the retrieved pages and pull them separately.
            page_num = 1

            for toc in journal.toc:

                # Skip pages previously downloaded if specified
                if page_num < from_page:
                    log_handler.log.info('Skipping page {0}'.format(page_num))
                    page_num += 1
                    continue

                if toc.url:
                    log_handler.log.info('Processing content {0} of {1}'.format(page_num, len(journal.toc)))

                    page = self.page_crawler.retrieve_page(journal_id, toc.original_id, toc.url)
                    toc.set_page(page)

                    # Calculate percentage per page, to keep consumers updated
                    self.progress_update(((page_num / len(journal.toc)) * 80) + 10)

                    self._process_page(toc.page, local_journal_handler)
                    # Calculate percentage per page, to keep consumers updated
                    self.progress_update((((page_num + 0.5) / len(journal.toc)) * 80) + 10)

                page_num += 1
        else:
            log_handler.log.warning('Processing single page for {0}'.format(journal_id))

            # Handle single-page journals/articles that have all the content on the title page
            journal.single_page = True

            # Create a single new page and set with the title page html
            content_page = Page(journal_id=journal_id, original_id=journal_id, original_html=journal.original_html)

            # Process it as a normal page and add it to the ToC
            content_page = self._process_page(content_page, local_journal_handler, single=True)
            journal.add_single_page(content_page)

            # Calculate percentage per page, to keep consumers updated
            self.progress_update(90)

        log_handler.log.info('Completed {0}'.format(journal_id), extra={'journal_id': journal_id})
        return journal

    def _process_page(self, page: Page, local_journal_handler, single=False):

        # Process the page and associated pics and maps
        page = self.page_crawler.parse_page(page, single=single)

        # Apply HTML post-processing for local browsability, if enabled
        if self.postprocess_html:
            self.html_postprocessor.postprocess_page(page)

        # Save original source and newly formatted resources
        page.save_originals(local_journal_handler)
        # Clear resources loaded into the page
        page.clear_resources()
        return page



from itertools import takewhile

from requests.packages.urllib3.util.retry import Retry

from bikesanity.io_utils import log_handler as log_handler


class CustomRetry(Retry):

    def get_backoff_time(self):
        # Modification here - remove instant retry
        consecutive_errors_len = len(
            list(
                takewhile(lambda x: x.redirect_location is None, reversed(self.history))
            )
        )
        backoff_value = self.backoff_factor * (2 ** (consecutive_errors_len-1))
        log_handler.log.warn('Backing off retry for {0} seconds'.format(backoff_value))
        return min(self.BACKOFF_MAX, backoff_value)

    def increment(self, method=None, url=None, response=None, error=None, _pool=None, _stacktrace=None, ):
        log_handler.log.warn('Retrying through custom retrier')
        if response and response.status == 429:
            log_handler.log.error('Received 429 - off we go')
            raise RuntimeError('Have been rate limited')
        if response and response.status == 403:
            log_handler.log.error('Received 403 - off we go')
            raise RuntimeError('Got an unauthorized')

        return super().increment(method, url, response, error, _pool, _stacktrace)



import io
import os
from abc import ABC, abstractmethod

from bikesanity.io_utils import log_handler
from bikesanity.io_utils.plain_session import PlainSession
from bikesanity.io_utils.throttled import ThrottledSession


class BaseRetriever(ABC):

    def __init__(self, base_url):
        self.base_url = base_url

    @abstractmethod
    def retrieve_index(self, url, error_message):
        pass

    @abstractmethod
    def retrieve_page(self, url, original_id, error_message):
        pass

    @abstractmethod
    def retrieve_image_large(self, path, error_message):
        pass

    def retrieve_image_small(self, path, error_message='{0}'):
        return self.retrieve_image_large(path, error_message)

    def retrieve_js(self, path):
        pass

    def retrieve_css(self, path):
        pass


class DownloadingRetriever(BaseRetriever):

    BASE_URL = 'https://www.crazyguyonabike.com'

    def __init__(self):
        super().__init__(self.BASE_URL)
        self.session = self._get_working_session(throttled=True)
        self.fast_session = self._get_working_session(throttled=False)

    def _get_working_session(self, throttled=True):
        session = None
        while not session:
            try:
                log_handler.log.warning('Attempting to create new session')
                session = ThrottledSession() if throttled else PlainSession()
                session.connect_session()
            except Exception as exc:
                log_handler.log.warning('Session creation failure - retrying: {0}'.format(exc))
                session = None
        return session

    def retrieve_index(self, url, error_message="{0}"):
        return self._retrieve_page(url, error_message)

    def retrieve_page(self, url, original_id, error_message="{0}"):
        page, resolved_url = self._retrieve_page(url, error_message)
        return page

    def _retrieve_page(self, url, error_message):
        try:
            response = self.session.make_request(url)
            if response.status_code == 200:
                return response.content, response.url
            else:
                message = error_message.format(url) + str(response.status_code)
                log_handler.log.error(message)
                raise RuntimeError(message)
        except Exception as exc:
            log_handler.log.error(error_message.format(exc))
            raise

    def retrieve_image_large(self, path, error_message="{0}"):
        try:
            url = self.base_url + path
            return self.download_binary(self.fast_session, url)
        except Exception as exc:
            log_handler.log.warn(error_message.format(exc))

    def retrieve_js(self, path):
        url = self.base_url + '/javascript/' + path
        return self.download_binary(self.fast_session, url)

    def retrieve_css(self, path):
        url = self.base_url + '/css/' + path
        return self.download_binary(self.fast_session, url)

    def download_binary(self, session, url):
        try:
            log_handler.log.info("Downloading {0}".format(url))
            with session.make_stream_request(url) as response:
                bin_bytes = io.BytesIO(response.raw.read())
            return bin_bytes
        except Exception as exc:
            log_handler.log.error("Failure to download: {0}".format(url))
            raise




class LocalRetriever(BaseRetriever):

    def __init__(self, local_handler):
        super().__init__('')
        self.local_handler = local_handler

    def retrieve_index(self, url, error_message="{0}"):
        return self.local_handler.get_index(), self.local_handler.journal_id

    def retrieve_page(self, url, original_id, error_message="{0}"):
        try:
            return self.local_handler.get_html_page_id(original_id)
        except Exception as exc:
            log_handler.log.error(error_message.format(exc))

    def retrieve_image_large(self, path, error_message="{0}"):
        try:
            # Remove any trailing querystrings from the url, or leading slashes
            if path.startswith('/'): path = path[1:]
            if '?' in path: path = path[:path.rfind('?')]

            return self.local_handler.get_image_binary(path)
        except Exception as exc:
            log_handler.log.error(error_message.format(exc))


class ExportRetriever(BaseRetriever):

    def __init__(self, local_handler):
        super().__init__('')
        self.local_handler = local_handler

    def retrieve_index(self, url, error_message='{0}'):
        try:
            return self.local_handler.get_html_page_id('index', additional_path='small'), self.local_handler.journal_id
        except Exception as exc:
            log_handler.log.error(error_message.format(exc))

    def retrieve_page(self, url, original_id, error_message='{0}'):
        try:
            return self.local_handler.get_html_page_id(original_id, additional_path='small/page')
        except Exception as exc:
            log_handler.log.error(error_message.format(exc))

    def retrieve_image_large(self, path, error_message='{0}'):
        try:
            # Take just the filename
            filename = os.path.basename(path)
            return self.local_handler.get_image_binary(filename, additional_path='large/pics')
        except Exception as exc:
            log_handler.log.error(error_message.format(exc))

    def retrieve_image_small(self, path, error_message='{0}'):
        try:
            # Take just the filename
            filename = os.path.basename(path)
            return self.local_handler.get_image_binary(filename, additional_path='small/pics')
        except Exception as exc:
            log_handler.log.error(error_message.format(exc))



import jsonpickle

from bikesanity.entities.content_blocks import Image, Map, TextBlock
from bikesanity.entities.journal import Journal
from bikesanity.entities.page import Page
from bikesanity.entities.json.journal import JsonJournal, JsonPage, JsonDistance, JsonSection, SectionType


class EnumHandler(jsonpickle.handlers.BaseHandler):
    def flatten(self, obj, data):  # data contains {}
        if obj == SectionType.TEXT:
            return 'text'
        elif obj == SectionType.PIC:
            return 'pic'

jsonpickle.handlers.registry.register(SectionType, EnumHandler)


class JsonOutput:

    def __init__(self, local_handler, progress_callback=None):
        self.local_handler = local_handler
        self.progress_callback = progress_callback

    def progress_update(self, percent):
        if self.progress_callback:
            self.progress_callback(progress=percent)

    def miles_primary(self, ds):
        return 'km)' in ds

    def miles_distance(self, ds):
        if not ds or ' miles' not in ds:
            return None
        return ds[:ds.find(' miles')] if self.miles_primary(ds) else ds[ds.find('(')+1:ds.find(' miles')]

    def km_distance(self, ds):
        if not ds or ' km' not in ds:
            return None
        return ds[ds.find('(')+1:ds.find(' km')] if self.miles_primary(ds) else ds[:ds.find(' km')]

    def journal_distance(self, journal: Journal):
        ds = journal.distance_statement or ''
        first_date, last_date = None, None

        if 'to' in ds:
            last_date = ds[ds.rfind('to')+3:]
        if 'on' in ds:
            first_date = ds[ds.rfind('on')+3:]
            last_date = ds[ds.rfind('on')+3:]
        if 'from' and 'to' in ds:
            first_date = ds[ds.rfind('from')+5:ds.rfind('to')]

        return self.miles_primary(ds), self.miles_distance(ds), self.km_distance(ds), first_date, last_date

    def output_journal(self, journal: Journal):

        json_journal = JsonJournal(
            doc_id=journal.journal_id,
            title=journal.journal_title,
            subtitle=journal.journal_subtitle
        )

        json_journal.add_author(journal.journal_author)

        miles_primary, miles, km, first_date, last_date = self.journal_distance(journal)
        json_journal.primaryDistanceUomIsMiles = miles_primary
        json_journal.distanceInMiles = miles
        json_journal.distanceInKilometers = km
        json_journal.firstDate = first_date
        json_journal.lastDate = last_date

        # Update progress
        self.progress_update(percent=10)

        # Iterate over the ToC and process every page
        page_idx = 1

        toc_pages = [toc for toc in journal.toc if toc.page]
        last_toc = toc_pages[-1]

        for toc_item in journal.toc:

            if toc_item.page:
                json_page = JsonPage(id=toc_item.original_id, url=toc_item.url)
                self.output_page(json_page, toc_item.page, page_idx, last=toc_item == last_toc)
                json_journal.add_page(json_page)
                page_idx += 1

            elif toc_item.title:
                json_page = JsonPage(id=None, url=None)
                json_page.contents.set_header(toc_item.title)


            # Calculate and update progress
            self.progress_update(((page_idx / len(toc_pages)) * 80) + 10)


        # Serialize to JSON
        json = jsonpickle.encode(json_journal, unpicklable=False, indent=4)
        self.local_handler.save_generated_json(json)

        # Update progress
        self.progress_update(percent=100)


    def output_page(self, json_page: JsonPage, page: Page, page_idx: int, last=False):

        json_page.contents.set_page(page.title, page.date_statement)

        json_page.contents.set_distance(
            JsonDistance(
                miles=self.miles_distance(page.page_distance),
                km=self.km_distance(page.page_distance),
                totalMilesSoFar=self.miles_distance(page.total_distance),
                totalKmSoFar=self.km_distance(page.total_distance)
            )
        )

        for content in page.contents:
            if isinstance(content, TextBlock):
                for para in content.content_text:
                    para = para.strip()
                    if para:
                        json_page.contents.add_section(
                            JsonSection.text_section(para)
                        )
            elif isinstance(content, Image):
                json_page.contents.add_section(
                    JsonSection.pic_section(
                        id=content.image_id,
                        url=content.original_path_fullsize,
                        filename='{0}.{1}'.format(content.image_id, content.extension),
                        caption=content.caption
                    )
                )
            elif isinstance(content, Map):
                # Maps not supported currently
                pass




from bs4 import BeautifulSoup
from io_utils.local_journal import LocalJournalHandler

from bikesanity.entities.content_blocks import Image, Map, TextBlock
from bikesanity.entities.journal import Journal
from bikesanity.entities.page import Page


class DiskHtmlOutput:

    OUTPUT_DIRECTORY = 'output/'

    def __init__(self, base_directory, journal_id):
        self.journal_id = journal_id
        self.base_directory = base_directory
        self.local_handler = LocalJournalHandler(base_directory, journal_id)

    def process_journal(self):
        # Remove existing files
        self.local_handler.remove_directory(self.OUTPUT_DIRECTORY)

        journal = self.local_handler.load_serialized_journal()

        if journal:
            self.output_journal(journal)

    def output_html(self, soup, filename):
        html = str(soup)
        filename = '{0}.html'.format(filename)
        path = self.local_handler.get_path(self.OUTPUT_DIRECTORY + filename)
        self.local_handler.output_text_to_file(path, html)
        return filename

    def output_journal(self, journal: Journal):
        soup = BeautifulSoup('', 'html5lib')

        body = soup.find('body')

        # Add title and distance statement
        self.output_title(soup, body, journal.journal_title)

        self.output_subtitle(soup, body, journal.journal_subtitle)
        self.output_subtitle(soup, body, 'By {0}'.format(journal.journal_author))

        locations = [location for location in journal.locales]
        self.output_para(soup, body, 'Locations: {0}'.format(', '.join(locations)))

        if journal.cover_image:
            self.output_picture(soup, body, journal.cover_image)

        toc_div = soup.new_tag('div', attrs={
            'class': 'toc_container'
        })

        # Iterate over the ToC and process every page
        page_idx = 1
        for toc_item in journal.toc:
            p_tag = soup.new_tag('p')
            toc_tag = p_tag

            if toc_item.page:
                html_filename = self.output_page(toc_item.page, page_idx)
                toc_tag = soup.new_tag('a', attrs={
                    'href': html_filename
                })
                p_tag.append(toc_tag)
                toc_tag.append(soup.new_string(toc_item.page.title))

                page_idx += 1
            elif toc_item.subtitle:
                toc_tag.append(soup.new_string(toc_item.subtitle))
            toc_div.append(p_tag)

        body.append(toc_div)
        self.output_html(soup, 'index')

    def output_page(self, page: Page, page_idx: int):
        soup = BeautifulSoup('', 'html5lib')

        body = soup.find('body')

        # Add title and distance statement
        h2_tag = soup.new_tag('h2')
        h2_tag.append(soup.new_string(page.title))
        body.append(h2_tag)

        h4_tag = soup.new_tag('h4')
        h4_tag.append(soup.new_string('Date: {0}'.format(page.date_statement)))
        body.append(h4_tag)
        h4_tag = soup.new_tag('h4')
        h4_tag.append(soup.new_string('Distance: {0}'.format(page.page_distance)))
        body.append(h4_tag)
        h4_tag = soup.new_tag('h4')
        h4_tag.append(soup.new_string('Total distance: {0}'.format(page.total_distance)))
        body.append(h4_tag)

        for content in page.contents:
            if isinstance(content, TextBlock):
                for para in content.content_text:
                    # Split every paragraph by explicit newlines
                    for p in para.split('\n'):
                        p_tag = soup.new_tag('p')
                        p_tag.append(soup.new_string(p))
                        body.append(p_tag)
            elif isinstance(content, Image):
                self.output_picture(soup, body, content)
            elif isinstance(content, Map):
                gpx_url = self.output_map(content)

                div_tag = soup.new_tag('div', attrs={
                    'class': 'map_container'
                })
                a_tag = soup.new_tag('a', attrs= {
                    'href': gpx_url
                })
                a_tag.append(soup.new_string("Map GPX"))
                div_tag.append(a_tag)

                if content.caption:
                    caption_tag = soup.new_tag('p')
                    caption_tag.append(soup.new_string(content.caption))
                    div_tag.append(caption_tag)

                body.append(div_tag)


        page_filename = self.output_html(soup, page_idx)
        return page_filename

    def output_picture(self, soup, body, image: Image):
        url_small, url_fullsize = self.output_pic(image)

        div_tag = soup.new_tag('div', attrs={
            'class': 'pic_container'
        })
        a_tag = soup.new_tag('a', attrs={
            'href': url_fullsize
        })
        img_tag = soup.new_tag('img', attrs={
            'src': url_small
        })
        a_tag.append(img_tag)
        div_tag.append(a_tag)

        if image.caption:
            caption_tag = soup.new_tag('p')
            caption_tag.append(soup.new_string(image.caption))
            div_tag.append(caption_tag)

        body.append(div_tag)

    def output_pic(self, image: Image):
        return '../resources/{0}.{1}.{2}'.format(image.image_id, 'small', image.extension), \
               '../resources/{0}.{1}'.format(image.image_id, image.extension)

    def output_map(self, map: Map):
        return '../resources/{0}.{1}'.format(map.map_id, 'gpx')

    def output_title(self, soup, body, text):
        h2_tag = soup.new_tag('h2')
        h2_tag.append(soup.new_string(text))
        body.append(h2_tag)

    def output_subtitle(self, soup, body, text):
        h4_tag = soup.new_tag('h4')
        h4_tag.append(soup.new_string(text))
        body.append(h4_tag)

    def output_para(self, soup, body, text):
        p_tag = soup.new_tag('p')
        p_tag.append(soup.new_string(text))
        body.append(p_tag)



import io
import os
import prettierfier

from bikesanity.entities.content_blocks import Image, Map
from .resources import get_resource_stream
from .serializer import Serializer
from .file_handler import FileHandler


class LocalJournalHandler(FileHandler):

    def __init__(self, base_path, journal_id):
        super().__init__()
        self.base_path = base_path
        self.journal_id = journal_id
        self.serializer = Serializer()

    def get_path(self, filename):
        return os.path.join(self.base_path, self.journal_id, filename)

    def get_base_path(self):
        return self.get_path('')

    def get_resource_path(self, filename):
        return os.path.join(self.base_path, self.journal_id, 'resources', filename)

    def get_html_path(self, filename):
        return os.path.join(self.base_path, self.journal_id, 'html', filename)

    def get_map_js_path(self, filename):
        return os.path.join(self.base_path, self.journal_id, 'html', 'maps', filename)

    def get_pdf_path(self, filename):
        return os.path.join(self.base_path, self.journal_id, 'pdf', filename)

    def get_index(self):
        return self.get_file_content(
            os.path.join(self.base_path, self.journal_id, 'index.html')
        )

    def get_html_page_id(self, page_id, additional_path=None):
        terms = [self.base_path, self.journal_id]
        if additional_path: terms.append(additional_path)
        terms.append(page_id + '.html')
        return self.get_file_content(os.path.join(*terms))

    def get_image_binary(self, path, additional_path=None):
        terms = [self.base_path, self.journal_id]
        if additional_path: terms.append(additional_path)
        terms.append(path)
        return io.BytesIO(self.get_binary_content(os.path.join(*terms)))

    def _resource_key(self, id, extension, suffix=''):
        return '{0}{1}.{2}'.format(id, suffix, extension)

    def save_html_original(self, filename, html):
        self.output_bytes_to_file(self.get_path(filename), html)

    def save_image_original(self, image: Image):
        if image.image_fullsize:
            self.output_binary_to_file(self.get_path(image.original_path_fullsize), image.image_fullsize)
        if image.image_small:
            self.output_binary_to_file(self.get_path(image.original_path_small), image.image_small)

    def save_map_original(self, map: Map):
        if map.gpx_data:
            self.output_binary_to_file(os.path.join(
                self.base_path, self.journal_id, 'maps', self._resource_key(map.map_id, 'gpx')
            ), map.gpx_data)
        if map.json_data:
            self.output_binary_to_file(os.path.join(
                self.base_path, self.journal_id, 'maps', self._resource_key(map.map_id, 'json')
            ), map.json_data)

    def save_image_resource(self, image: Image):
        # Output fullsize image
        if image.image_fullsize:
            self.output_binary_to_file(os.path.join(
                self.base_path, self.journal_id, 'resources', self._resource_key(image.image_id, image.extension)
            ), image.image_fullsize)

        # Output small image
        if image.image_small:
            self.output_binary_to_file(os.path.join(
                self.base_path, self.journal_id, 'resources', self._resource_key(image.image_id, image.extension, suffix='.small')
            ), image.image_small)

    def save_map_resource(self, map: Map):
        if map.gpx_data:
            self.output_binary_to_file(os.path.join(
                self.base_path, self.journal_id, 'resources', self._resource_key(map.map_id, 'gpx')
            ), map.gpx_data)
        if map.json_data:
            self.output_binary_to_file(os.path.join(
                self.base_path, self.journal_id, 'resources', self._resource_key(map.map_id, 'json')
            ), map.json_data)

    def load_map_gpx(self, map: Map):
        return self.get_file_content(os.path.join(
            self.base_path, self.journal_id, 'resources', self._resource_key(map.map_id, 'gpx')
        ))

    def save_js_resource(self, js, content):
        if content:
            self.output_binary_to_file(os.path.join(
                self.base_path, self.journal_id, 'javascript', js
            ), content)

    def save_css_resource(self, css, content):
        if content:
            self.output_binary_to_file(os.path.join(
                self.base_path, self.journal_id, 'css', css
            ), content)

    def save_generated_html(self, soup, filename):
        html = prettierfier.prettify_html(soup.prettify())
        filename = '{0}.html'.format(filename)
        path = self.get_html_path(filename)
        self.output_text_to_file(path, html)
        return filename

    def save_map_js(self, map_id, map_js):
        path = self.get_map_js_path('{0}.js'.format(map_id))
        self.output_text_to_file(path, map_js)

    def save_generated_json(self, json):
        path = self.get_path('journal.json')
        self.output_text_to_file(path, json)

    def save_generated_pdf(self, pdf, part=None):
        pdf_filename = 'journal.pdf' if not part else 'journal_part_{0}.pdf'.format(part)
        path = self.get_pdf_path(pdf_filename)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        pdf.output(path)


    def serialize_and_save_journal(self, journal):
        serialized = self.serializer.serialize_journal(journal)
        self.output_binary_to_file(os.path.join(
            self.base_path, self.journal_id, 'journal.pickle'
        ), serialized)

    def load_serialized_journal(self):
        return self.serializer.unserialize_from_data(
            self.get_binary(os.path.join(
                self.base_path, self.journal_id, 'journal.pickle'
            ))
        )

    def remove_directory(self, path):
        super().remove_directory(os.path.join(
            self.base_path, self.journal_id, path
        ))

    def copy_resource_stream(self, paths, output_paths):
        stream = get_resource_stream(paths)

        full_output_paths = [ self.base_path, self.journal_id, 'html' ]
        full_output_paths.extend(output_paths)

        self.output_binary_to_file(os.path.join(*full_output_paths), stream)
        stream.close()



from typing import Dict
import json
from io import BytesIO
from bs4 import BeautifulSoup
import gpxpy
import gpxpy.gpx



class MapExtractor:
    def __init__(self, elem=None):
        self.docs = []
        if elem: self.docs.append(elem)

    def add_doc(self, html):
        self.docs.append(BeautifulSoup(html, features="lxml"))

    def iterate_map_data(self):
        for doc in self.docs:
            map_data_elems = doc.findAll("div", {"id" : lambda l: l and l.startswith('map') and l.endswith('data')})
            for map_data_elem in map_data_elems:
                yield map_data_elem.text


class LineExtractor:
    def __init__(self, map_data: str):
        self.raw_data = map_data
        self.map_data = json.loads(map_data)
        self.line = self._get_line(self.map_data)

    def _get_line(self, map_data) -> Dict:
        return map_data['lines'][0] if 'lines' in map_data and len(map_data['lines']) > 0 else None

    def get_map_id(self):
        return str(self.line.get('line_id', None)) if self.line else None

    def generate_track_segment(self, polyline: Dict):
        gpx_segment = gpxpy.gpx.GPXTrackSegment()

        # Add all the points from this poorly thought-out data structure
        if 'lngs' not in polyline or 'lats' not in polyline or len(polyline['lngs']) != len(polyline['lats']): return None
        for lat, long in zip(polyline['lats'], polyline['lngs']):
            point = gpxpy.gpx.GPXTrackPoint(latitude=lat, longitude=long)
            gpx_segment.points.append(point)

        return gpx_segment

    def serialize_gpx(self, gpx: gpxpy.gpx.GPX):
        return gpx.to_xml()

    def generate_gpx(self):
        gpx = gpxpy.gpx.GPX()

        # Create a single track
        gpx_track = gpxpy.gpx.GPXTrack()
        gpx_track.name = 'Track'
        gpx.tracks.append(gpx_track)

        if 'polylines' not in self.line: return gpx

        # Iterate over all the "polylines", and add them as separate track segment
        for polyline in self.line['polylines']:
            track_segement = self.generate_track_segment(polyline)
            if track_segement: gpx_track.segments.append(track_segement)

        return self.serialize_gpx(gpx)

    def generate_map_binary_data(self):
        gpx_xml = self.generate_gpx()
        return BytesIO(gpx_xml.encode()) if gpx_xml else None, BytesIO(self.raw_data.encode()) if self.raw_data else None



"""
This is the main entry point for the command-line runner
"""

import os
from pathlib import Path
import click

from bikesanity.io_utils.log_handler import init_logging, log

from bikesanity.processing.download_journal import DownloadJournal
from bikesanity.processing.load_disk_journal import LoadDiskJournal
from bikesanity.processing.publish_journal import PublishJournal, PublicationFormats


BASE_DIRECTORY = 'CycleSanityJournals'

base_path = os.path.join(Path.home(), BASE_DIRECTORY)

@click.group()
def run():
    init_logging()
    os.makedirs(base_path, exist_ok=True)


@run.command()
@click.argument('journal_link')
@click.option('--do-process', is_flag=True, help="Interpret journal content and process resources")
@click.option('--location', default=None, help="Custom download location for journal files")
@click.option('--no-local-readability-postprocessing', is_flag=True, help="Interpret journal content and process resources")
@click.option('--from-page', default=0, help="Pick up a download from the specified page")
def download(journal_link, do_process=False, location=None, no_local_readability_postprocessing=False, from_page=0):
    log.info('Starting download task...')

    download_path = location if location else base_path
    journal = None

    # Download the journal
    try:
        journal_downloader = DownloadJournal(download_path, postprocess_html=not no_local_readability_postprocessing)
        journal = journal_downloader.download_journal_url(journal_link, from_page)

        log.info('Completed download task! Journal downloaded to {0}'.format(journal_downloader.get_download_location(journal)))
    except Exception:
        log.exception('Critical error on downloading journal')

    if do_process and journal:
        process(journal.journal_id, location)


@run.command()
@click.argument('journal_id')
@click.option('--input-location', default=None, help="Custom download location of journal")
@click.option('--output-location', default=None, help="Custom output for processed journals")
def process(journal_id, input_location=None, output_location=None):
    log.info('Processing journal id {0}'.format(journal_id))

    input_path = input_location if input_location else base_path
    output_path = output_location if output_location else base_path

    try:
        journal_processor = LoadDiskJournal(input_path, output_path, journal_id, exported=False)
        journal = journal_processor.load_journal_from_disk()
        log.info('Completed processing task! Processed journal available in {0}'.format(journal_processor.get_process_location()))

    except Exception:
        log.exception('Critical error on processing journal')



@run.command()
@click.argument('journal_id')
@click.option('--location', help="Custom location of exported journal")
@click.option('--output-location', default=None, help="Custom output for processed journals")
def process_exported(journal_id, location, output_location):
    log.info('Processing previously exported journal {0}'.format(journal_id))

    input_path = location if location else base_path
    output_path = output_location if output_location else base_path

    try:
        journal_processor = LoadDiskJournal(input_path, output_path, journal_id, exported=True)
        journal = journal_processor.load_journal_from_disk()
        log.info('Completed processing task! Processed journal available in {0}'.format(journal_processor.get_process_location()))

    except Exception:
        log.exception('Critical error on processing journal')


@run.command()
@click.argument('journal_id')
@click.option('--input-location', default=None, help="Custom download location of journal")
@click.option('--output-location', default=None, help="Custom output for processed journals")
@click.option('--html', is_flag=True, default=False, help="Export as HTML")
@click.option('--json', is_flag=True, default=False, help="Export as JSON")
@click.option('--pdf', is_flag=True, default=False, help="Export as PDF")
@click.option('--epub', is_flag=True, default=False, help="Export as EPUB")
def publish(journal_id, input_location, output_location, html, json, pdf, epub):
    input_path = input_location if input_location else base_path
    output_path = output_location if output_location else base_path

    if not html and not json and not pdf and not epub: html = True

    formats = []
    if html: formats.append('html')
    if json: formats.append('json')
    if pdf: formats.append('pdf')

    log.info('Outputting journal id {0} to formats: {1}'.format(journal_id, ', '.join(formats)))

    try:
        journal_publisher = PublishJournal(input_path, output_path, journal_id)

        if html:
            journal_publisher.publish_journal_id(PublicationFormats.TEMPLATED_HTML)
            log.info('Completed publishing to HTML! Published journal available in {0}'.format(journal_publisher.get_publication_location()))
        if json:
            journal_publisher.publish_journal_id(PublicationFormats.JSON_MODEL)
            log.info('Completed publishing to JSON! Published journal available in {0}'.format(journal_publisher.get_publication_location()))
        if pdf:
            journal_publisher.publish_journal_id(PublicationFormats.PDF)
            log.info('Completed publishing to PDF! Published journal available in {0}'.format(journal_publisher.get_publication_location()))

    except Exception:
        log.exception('Critical error on publishing journal')

@run.command()
def version():
    print('BikeSanity script v1.1.0')


if __name__ == '__main__':
    run()



import bikesanity.io_utils.log_handler as log_handler
from bikesanity.entities.content_blocks import Image, Map, TextBlock
from bikesanity.entities.journal import Journal

from .journal_pdf import JournalPdf


class PdfTocEntry:
    def __init__(self, title, page_no, is_header):
        self.is_header = is_header
        self.page_no = str(page_no)
        self.title = title


class PdfOutput:

    MAX_PDF_SECTIONS = 25


    def __init__(self, local_handler, progress_callback=None):
        self.local_handler = local_handler
        self.progress_callback = progress_callback

    def progress_update(self, percent):
        if self.progress_callback:
            self.progress_callback(progress=percent)


    def chunks(self, lst, n):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    def output_journal(self, journal: Journal):

        # Determine if it needs to be split into parts, and make the split
        full_pages = [content for content in journal.toc if content.page]
        part_progress = 100 / (len(full_pages)/self.MAX_PDF_SECTIONS)

        if len(full_pages) <= self.MAX_PDF_SECTIONS:
            log_handler.log.info('Writing to single PDF file')
            self.output_journal_part(journal, journal.toc, part_progress=100)
        else:
            progress = 0
            start_page = 0
            part = 1

            log_handler.log.info('Splitting up into multiple PDF files')
            for chunk in self.chunks(journal.toc, self.MAX_PDF_SECTIONS):
                log_handler.log.info('Creating PDF part {0}'.format(part))
                self.output_journal_part(journal, chunk, part=part, from_page=start_page, start_progress=progress, part_progress=part_progress)
                part += 1
                start_page += self.MAX_PDF_SECTIONS
                progress += part_progress

        # Update progress
        self.progress_update(percent=100)


    def output_journal_part(self, journal: Journal, contents, part=None, from_page=None, start_progress=0, part_progress=100):
        journal_pdf = JournalPdf(title=journal.journal_title, author=journal.journal_author, part=part)

        # Create a cover
        journal_pdf.set_auto_page_break(False, margin=journal_pdf.MARGIN)
        journal_pdf.cover_title(journal.journal_title, journal.journal_subtitle, journal.journal_author, journal.distance_statement, part=part)

        journal_pdf.clipping_rect(0, 100, journal_pdf.A4_WIDTH, 180, False)
        if journal.cover_image:
            self.output_image(journal_pdf, journal.cover_image)
        journal_pdf.set_auto_page_break(True, margin=journal_pdf.MARGIN)
        journal_pdf.unset_clipping()

        # Leave space for the TOC
        journal_pdf.add_page()

        pdf_toc = []

        # Process pages
        page_idx = from_page or 0
        new_page_needed = True

        for toc_item in contents:

            if toc_item.page:
                log_handler.log.info('Processing page {0} to PDF'.format(page_idx))
                page_idx += 1

                journal_pdf.update_page_title(toc_item.page.title)

                if new_page_needed: journal_pdf.add_page()
                pdf_toc.append(PdfTocEntry(toc_item.page.title, journal_pdf.page_no()-2, False))

                journal_pdf.chapter_title(
                    toc_item.page.title, toc_item.page.date_statement,
                    toc_item.page.page_distance, toc_item.page.total_distance
                )

                for content in toc_item.page.contents:
                    if isinstance(content, TextBlock):
                        self.output_textblock(journal_pdf, content)
                    elif isinstance(content, Image):
                        self.output_image(journal_pdf, content)
                    elif isinstance(content, Map):
                        self.output_map(journal_pdf, content)

                new_page_needed = True

            elif toc_item.title:
                journal_pdf.update_page_title(toc_item.title)
                journal_pdf.add_page()
                journal_pdf.section_title(toc_item.title)
                pdf_toc.append(PdfTocEntry(toc_item.title, journal_pdf.page_no()-2, True))
                new_page_needed = False

            # Calculate and update progress
            self.progress_update(((page_idx / len(contents)) * (part_progress/2)) + start_progress)

        # Go back and populate the table of contents
        final_page = journal_pdf.page_no()
        journal_pdf.page = 2
        journal_pdf.set_y(30)

        journal_pdf.add_toc(pdf_toc)

        # Reset to the last page
        journal_pdf.page = final_page

        log_handler.log.info('Writing PDF file (this can be CPU intensive and take a few minutes)')
        self.progress_update(start_progress + (part_progress/1.5))
        self.local_handler.save_generated_pdf(journal_pdf, part=part)

        self.progress_update(start_progress + part_progress)

        # Cleanup font files
        journal_pdf.cleanup_tmp_files()


    def output_textblock(self, journal_pdf: JournalPdf, content: TextBlock):
        for para in content.content_text:
            para = para.strip()

            if not para: continue

            # Don't write out the previous page section dividers, as they're no longer needed
            if para.startswith('>>>') or para.startswith('<<<'): continue

            journal_pdf.text_content(para)

    def output_image(self, journal_pdf: JournalPdf, image: Image, height=None):
        url_fullsize = self.output_pic(image)
        journal_pdf.add_image(url_fullsize, image.caption, height=height)

    def output_map(self, journal_pdf: JournalPdf, map: Map):
        pass

    def output_pic(self, image: Image):
        return self.local_handler.get_resource_path('{0}.{1}'.format(image.image_id, image.extension))



from .content_blocks import Image
from .page import Page


class TocEntry:
    def __init__(self, original_id, title, url):
        self.original_id = original_id
        self.title = title
        self.url = url
        self.page = None

    def set_page(self, page):
        self.page = page


class Journal:

    def __init__(self, journal_id, original_html):
        self.journal_id = journal_id

        self.original_html = original_html
        self.postprocessed_html = None

        self.journal_title = None
        self.journal_subtitle = None
        self.journal_author = None
        self.distance_statement = None
        self.locales = []

        self.cover_image = None

        self.toc = []
        self.single_page = False

        self.js = {}
        self.css = {}

    def add_toc_entry(self, original_id, title, url):
        toc_entry = TocEntry(original_id, title, url)
        self.toc.append(toc_entry)

    def add_single_page(self, page: Page):
        toc_entry = TocEntry(self.journal_id, self.journal_title, None)
        toc_entry.set_page(page)
        self.toc.append(toc_entry)

    def add_cover_image(self, image: Image):
        self.cover_image = image

    def save_original_source(self, local_handler):
        # Save the HTML itself
        local_handler.save_html_original('index.html', self.postprocessed_html if self.postprocessed_html else self.original_html)

        # Save the cover image if one exists
        if self.cover_image: local_handler.save_image_original(self.cover_image)

        # Save the JS and CSS
        for js, content in self.js.items():
            local_handler.save_js_resource(js, content)
        for css, content in self.css.items():
            local_handler.save_css_resource(css, content)

    def save_resources(self, local_handler):
        # Save the cover image if one exists
        if self.cover_image: local_handler.save_image_resource(self.cover_image)

    def clear_resources(self):
        if self.cover_image: self.cover_image.clear_resources()

        for js in self.js.values():
            if js:
                js.close()
        for css in self.css.values():
            if css:
                css.close()

    def update_data_model(self):
        if not hasattr(self, 'single_page'): self.single_page = False



from enum import Enum, auto
import os

import bikesanity.io_utils.log_handler as log_handler
from bikesanity.entities.journal import Journal
from bikesanity.io_utils.local_journal import LocalJournalHandler

from bikesanity.output_transformers.templated_html_output import TemplatedHtmlOutput
from bikesanity.output_transformers.json_output import JsonOutput
from bikesanity.output_transformers.pdf_output import PdfOutput


class PublicationFormats(Enum):
    TEMPLATED_HTML = auto()
    JSON_MODEL = auto()
    PDF = auto()


class PublishJournal:

    PROCESSED_DIRECTORY = 'processed'
    TEMPLATES_DIRECTORY = '../templates/'

    def __init__(self, input_location, output_location, journal_id):
        self.journal_id = journal_id

        self.input_location = os.path.join(input_location, self.PROCESSED_DIRECTORY)
        self.output_location = os.path.join(output_location, self.PROCESSED_DIRECTORY)

        self.input_handler = LocalJournalHandler(self.input_location, journal_id)
        self.output_handler = LocalJournalHandler(self.output_location, journal_id)

        self.publication_location = None

    def get_publication_location(self):
        return self.publication_location

    def publish_journal_id(self, format: PublicationFormats, progress_callback=None):
        # Load the journal by unpickling from the processed form
        journal = self.input_handler.load_serialized_journal()
        if not journal:
            log_handler.log.error('Failure to load any journal with ID {0} - have you processed it?'.format(self.journal_id))
            return

        # Switch based on journal format required
        if format == PublicationFormats.TEMPLATED_HTML:
            self.publish_journal_templated_html(journal, progress_callback)
        elif format == PublicationFormats.JSON_MODEL:
            self.export_json_model(journal, progress_callback)
        elif format == PublicationFormats.PDF:
            self.publish_pdf(journal, progress_callback)


    def publish_journal_templated_html(self, journal: Journal, progress_callback=None):
        templated_output = TemplatedHtmlOutput(self.output_handler, progress_callback)
        templated_output.output_journal(journal)
        self.publication_location = self.output_handler.get_html_path('index.html')

    def export_json_model(self, journal: Journal, progress_callback=None):
        json_output = JsonOutput(self.output_handler, progress_callback)
        json_output.output_journal(journal)
        self.publication_location = self.output_handler.get_path('journal.json')

    def publish_pdf(self, journal: Journal, progress_callback=None):
        pdf_output = PdfOutput(self.output_handler, progress_callback)
        pdf_output.output_journal(journal)
        self.publication_location = self.output_handler.get_pdf_path('journal.pdf')



import os
import bikesanity.io_utils.log_handler as log_handler
from bikesanity.entities.journal import Journal
from bikesanity.entities.page import Page
from bikesanity.io_utils.local_journal import LocalJournalHandler
from bikesanity.services.retrievers import LocalRetriever, ExportRetriever

from bikesanity.interpreter.journal_content import JournalContent
from bikesanity.interpreter.page_interpreter import PageInterpreter


class LoadDiskJournal:

    DOWNLOAD_DIRECTORY = 'downloads'
    EXPORTED_DIRECTORY = 'exported'
    PROCESSED_DIRECTORY = 'processed'

    def __init__(self, input_location, output_location, journal_id, exported=False, progress_callback=None):

        self.input_location = os.path.join(input_location, self.EXPORTED_DIRECTORY if exported else self.DOWNLOAD_DIRECTORY)
        self.output_location = os.path.join(output_location, self.PROCESSED_DIRECTORY)
        self.progress_callback = progress_callback

        self.journal_id = journal_id
        self.input_handler = LocalJournalHandler(self.input_location, journal_id)
        self.output_handler = LocalJournalHandler(self.output_location, journal_id)

        # Ensure the output is clear
        self.output_handler.remove_directory('')
        os.makedirs(self.output_location, exist_ok=True)

        self.retriever = ExportRetriever(self.input_handler) if exported else LocalRetriever(self.input_handler)
        self.outputter = LocalRetriever(self.output_handler)

        self.journal_crawler = JournalContent(self.retriever)
        self.page_crawler = PageInterpreter(self.retriever)

    def progress_update(self, percent):
        if self.progress_callback:
            self.progress_callback(progress=percent)

    def get_process_location(self):
        return self.output_handler.get_base_path()

    def load_journal_from_disk(self):
        # Retrieve and process the journal content
        journal = self.journal_crawler.retrieve_journal(None, self.journal_id)
        self.progress_update(percent=10)
        journal = self._process_journal(journal)
        return journal

    def _process_journal(self, journal: Journal):
        journal_id = journal.journal_id

        # Handle standard multi-page journals with a ToC
        if journal.toc:
            log_handler.log.warning('Processing multiple pages for {0}'.format(journal_id))

            # Iterate over all the retrieved pages and pull them separately.
            page_count = 0
            for toc in journal.toc:
                if toc.url:
                    page = self._process_page(toc.original_id)
                    toc.set_page(page)

                # Calculate percentage per page, to keep consumers updated
                self.progress_update(((page_count / len(journal.toc)) * 80) + 10)
                page_count += 1

        else:
            log_handler.log.warning('Processing single page for {0}'.format(journal_id))

            # Handle single-page journals/articles that have all the content on the title page
            journal.single_page = True

            # Create a single new page and set with the title page html
            content_page = Page(journal_id=journal_id, original_id=journal_id, original_html=journal.original_html)

            # Process it as a normal page and add it to the ToC
            content_page = self._process_page(content_page, single=True)
            journal.add_single_page(content_page)
            self.progress_update(percent=90)

        # Save and clear any resources not associated with pages
        journal.save_resources(self.output_handler)
        journal.clear_resources()

        # Finally serialize the parsed data structure and output
        log_handler.log.info('Serializing data for {0}'.format(journal_id), extra={'journal_id': journal_id})
        self.output_handler.serialize_and_save_journal(journal)

        self.progress_update(percent=100)
        log_handler.log.info('Completed {0}'.format(journal_id), extra={'journal_id': journal_id})
        return journal

    def _process_page(self, page_id, single=False):
        log_handler.log.warning('Processing page {0} for {1}'.format(page_id, self.journal_id))

        # Process the page and associated pics and maps
        page = self.page_crawler.retrieve_page(self.journal_id, page_id, None)
        page = self.page_crawler.parse_page(page, single=single)

        # Save locally and clear resources loaded into the page
        page.save_resources(self.output_handler)
        page.clear_resources()

        return page



import re

from bs4 import BeautifulSoup

from bikesanity.entities.content_blocks import Image, Map, TextBlock
from bikesanity.entities.journal import Journal
from bikesanity.entities.page import Page
from bikesanity.io_utils.resources import get_resource_string


class TemplatedHtmlOutput:

    TEMPLATES_DIRECTORY = 'templates'

    def __init__(self, local_handler, progress_callback=None):
        self.local_handler = local_handler
        self.progress_callback = progress_callback

    def progress_update(self, percent):
        if self.progress_callback:
            self.progress_callback(progress=percent)

    def open_index_template(self):
        template_html = get_resource_string([self.TEMPLATES_DIRECTORY, 'index.html'])
        return BeautifulSoup(template_html, 'html5lib')

    def open_page_template(self):
        template_html = get_resource_string([self.TEMPLATES_DIRECTORY, 'page.html'])
        return BeautifulSoup(template_html, 'html5lib')


    def set_title(self, soup, title):
        soup.title.string.replace_with(title)

    def generate_logo(self, soup, author, journal_title):
        logo = soup.find(id='logo')
        if author:
            logo.append(soup.new_string(author))
        logo_span = soup.new_tag('span')
        if journal_title:
            logo_span.append(soup.new_string(' - {0}'.format(journal_title)))
        logo.append(logo_span)


    def output_journal(self, journal: Journal):
        soup = self.open_index_template()

        self.set_title(soup, journal.journal_title)

        self.generate_logo(soup, journal.journal_author, journal.journal_title)

        index_header = soup.find(id='index_header')

        # Add title and distance statement
        self.output_title(soup, index_header, journal.journal_title)

        self.output_subtitle(soup, index_header, journal.journal_subtitle)
        subtitle = self.output_subtitle(soup, index_header, 'By ')
        strong_tag = soup.new_tag('strong')
        if journal.journal_author:
            strong_tag.append(soup.new_string(journal.journal_author))
        subtitle.append(strong_tag)

        locations = [location for location in journal.locales]
        self.output_p(soup, index_header, 'Locations: {0}'.format(', '.join(locations)))

        index_body = soup.find(id='index_page')

        if journal.cover_image:
            self.output_picture(soup, index_body, journal.cover_image, fullsize_only=True, prepend=True)

        # Update progress
        self.progress_update(percent=10)

        toc_div = soup.find(id='toc_container')

        # Iterate over the ToC and process every page
        page_idx = 1
        ul_tag = None

        toc_pages = [toc for toc in journal.toc if toc.page]
        last_toc = toc_pages[-1]

        for toc_item in journal.toc:

            if toc_item.page:
                if not ul_tag:
                    ul_tag = soup.new_tag('ul')

                li_tag = soup.new_tag('li')

                html_filename = self.output_page(journal, toc_item.page, page_idx, last=toc_item == last_toc)
                a_tag = soup.new_tag('a', attrs={
                    'href': html_filename
                })
                a_tag.append(soup.new_string(toc_item.page.title))
                li_tag.append(a_tag)
                ul_tag.append(li_tag)

                page_idx += 1
            elif toc_item.title:
                if ul_tag:
                    toc_div.append(ul_tag)
                    ul_tag = None

                self.output_subtitle(soup, toc_div, toc_item.title)

            # Calculate and update progress
            self.progress_update(((page_idx / len(toc_pages)) * 80) + 10)

        if ul_tag:
            toc_div.append(ul_tag)

        # Save the generated HTML
        self.local_handler.save_generated_html(soup, 'index')

        # Copy the resources to the output as well
        self.local_handler.copy_resource_stream(['templates', 'js', 'maps.js'], ['js', 'maps.js'])
        self.local_handler.copy_resource_stream(['templates', 'css', 'journal.css'], ['css', 'journal.css'])
        self.local_handler.copy_resource_stream(['templates', 'css', 'responsive.css'], ['css', 'responsive.css'])
        self.local_handler.copy_resource_stream(['templates', 'css', 'fontawesome-all.min.css'], ['css', 'fontawesome-all.min.css'])
        self.local_handler.copy_resource_stream(['templates', 'css', 'images', 'dark-tl.svg'], ['css', 'images', 'dark-tl.svg'])
        self.local_handler.copy_resource_stream(['templates', 'css', 'images', 'dark-tr.svg'], ['css', 'images', 'dark-tr.svg'])
        self.local_handler.copy_resource_stream(['templates', 'css', 'images', 'dark-br.svg'], ['css', 'images', 'dark-br.svg'])
        self.local_handler.copy_resource_stream(['templates', 'css', 'images', 'dark-bl.svg'], ['css', 'images', 'dark-bl.svg'])

        # Update progress
        self.progress_update(percent=100)


    def add_nav_link(self, soup, nav_ul, idx, text):
        a_tag = soup.new_tag('a', attrs= {
         'href': '{0}.html'.format(idx)
        })
        a_tag.append(soup.new_string(text))
        li_tag = soup.new_tag('li')
        li_tag.append(a_tag)
        nav_ul.append(li_tag)


    def output_page(self, journal: Journal, page: Page, page_idx: int, last=False):
        soup = self.open_page_template()
        self.set_title(soup, page.title)

        self.generate_logo(soup, journal.journal_author, journal.journal_title)

        header = soup.find(id='header')
        nav_ul = header.find('ul')
        if page_idx > 1:
            self.add_nav_link(soup, nav_ul, page_idx-1, 'Previous')
        if not last:
            self.add_nav_link(soup, nav_ul, page_idx+1, 'Next')

        page_header = soup.find(id='page_header')

        title_words = page.title.split(' ')
        title_strong = ' '.join(title_words[:2] if len(title_words) > 2 else title_words)
        title_weak = ' '.join(title_words[2:] if len(title_words) > 2 else [])
        header_title = self.output_title(soup, page_header, title_weak)
        if title_strong:
            strong = soup.new_tag('strong')
            strong.append(soup.new_string(title_strong + ' '))
            header_title.insert(0, strong)

        page_div = soup.find(id='journal_page')

        # Add title and distance statement
        if page.date_statement:
            h4_tag = soup.new_tag('h4')
            strong_tag = soup.new_tag('strong')
            strong_tag.append(soup.new_string('Date: '))
            h4_tag.append(strong_tag)
            h4_tag.append(soup.new_string(page.date_statement))
            page_div.append(h4_tag)
        if page.page_distance:
            h4_tag = soup.new_tag('h4')
            strong_tag = soup.new_tag('strong')
            strong_tag.append(soup.new_string('Distance: '))
            h4_tag.append(strong_tag)
            h4_tag.append(soup.new_string(page.page_distance))
            page_div.append(h4_tag)
        if page.total_distance:
            h4_tag = soup.new_tag('h4')
            strong_tag = soup.new_tag('strong')
            strong_tag.append(soup.new_string('Total distance: '))
            h4_tag.append(strong_tag)
            h4_tag.append(soup.new_string(page.total_distance))
            page_div.append(h4_tag)

        for content in page.contents:
            if isinstance(content, TextBlock):
                for para in content.content_text:
                    self.output_para(soup, page_div, para)
            elif isinstance(content, Image):
                self.output_picture(soup, page_div, content)
            elif isinstance(content, Map):
                self.output_map(soup, page_div, content)

        page_filename = self.local_handler.save_generated_html(soup, page_idx)
        return page_filename

    def output_para(self, soup, page_div, para):
        p_tag = soup.new_tag('p')

        # Split every paragraph by explicit newlines
        first = True
        for p in para.split('\n'):
            # Write out a <br> not the first line
            if not first: p_tag.append(soup.new_tag('br'))

            # Don't write out the previous page section dividers, as they're no longer needed
            if p.startswith('>>>') or p.startswith('<<<'):
                continue

            # Links will be separated in their own paragraphs. Convert them into proper links
            if p.startswith('http://') or p.startswith('https://') and ' ' not in p:
                a_tag = soup.new_tag('a', attrs= {
                    'href': p
                })
                a_tag.append(soup.new_string(p))
                p_tag.append(a_tag)
            else:
                p_tag.append(soup.new_string(p))

            first = False

        page_div.append(p_tag)


    MAP_GPX_SCRIPT = "<!--\nvar {0} = '{1}';\n-->"

    MAP_SCRIPT = "\naddMap('{0}', {1});\n"

    def sanitize_gpx(self, gpx):
        gpx = gpx.replace('\n', '')
        return re.sub(r'\s{2,}', ' ', gpx)

    def create_map_js(self, map: Map, map_id):
        # Add the JavaScript for the map
        map_gpx = self.local_handler.load_map_gpx(map)
        map_gpx = self.sanitize_gpx(map_gpx)

        map_gpx_var = '{0}_gpx'.format(map_id).replace('-', '_')
        map_js = self.MAP_GPX_SCRIPT.format(map_gpx_var, map_gpx)
        self.local_handler.save_map_js(map_id, map_js)
        return map_gpx_var


    def output_map(self, soup, page_div, map: Map):
        map_container_div_tag = soup.new_tag('div', attrs={
            'class': 'map_container'
        })

        gpx_url = self.get_map_url(map)
        gpx_span_tag = soup.new_tag('span', attrs={
            'class': 'map_link'
        })
        gpx_a_tag = soup.new_tag('a', attrs= {
            'href': gpx_url
        })
        gpx_a_tag.append(soup.new_string("Map GPX"))
        gpx_span_tag.append(gpx_a_tag)
        map_container_div_tag.append(gpx_span_tag)

        map_id = 'map_{0}'.format(map.map_id)
        map_div = soup.new_tag('div', attrs={
            'id': map_id,
            'class': 'map'
        })
        map_container_div_tag.append(map_div)

        if map.caption:
            caption_tag = soup.new_tag('p')
            caption_tag.append(soup.new_string(map.caption))
            map_container_div_tag.append(caption_tag)


        # Create the map wrapped as a JS file, to keep the HTML clean
        map_gpx_var = self.create_map_js(map, map_id)
        map_js_tag = soup.new_tag('script', attrs={
            'src': 'maps/{0}.js'.format(map_id)
        })
        map_container_div_tag.append(map_js_tag)

        script_tag = soup.new_tag('script', attrs={
            'language': 'JavaScript',
            'type': 'text/javascript'
        })
        script = self.MAP_SCRIPT.format(map_id, map_gpx_var)
        script_tag.append(soup.new_string(script))
        map_container_div_tag.append(script_tag)

        page_div.append(map_container_div_tag)


    def output_picture(self, soup, body, image: Image, fullsize_only=False, prepend=False):
        url_small, url_fullsize = self.output_pic(image)

        div_tag = soup.new_tag('div', attrs={
            'class': 'pic_container'
        })
        wrapper = soup.new_tag('div', attrs={
            'class': 'image-wrapper'
        })
        a_tag = soup.new_tag('a', attrs={
            'href': url_fullsize
        })
        img_tag = soup.new_tag('img', attrs={
            'src': url_fullsize if fullsize_only else url_small
        })
        a_tag.append(img_tag)
        wrapper.append(a_tag)

        div_tag.append(wrapper)

        if image.caption:
            caption_tag = soup.new_tag('p')
            caption_tag.append(soup.new_string(image.caption))
            div_tag.append(caption_tag)

        if prepend:
            body.insert(0, div_tag)
        else:
            body.append(div_tag)

    def output_pic(self, image: Image):
        return '../resources/{0}.{1}.{2}'.format(image.image_id, 'small', image.extension), \
               '../resources/{0}.{1}'.format(image.image_id, image.extension)


    def get_map_url(self, map: Map):
        return '../resources/{0}.{1}'.format(map.map_id, 'gpx')

    def output_title(self, soup, body, text):
        h2_tag = soup.new_tag('h2')
        h2_tag.append(soup.new_string(text))
        body.append(h2_tag)
        return h2_tag

    def output_subtitle(self, soup, body, text):
        h4_tag = soup.new_tag('h4')
        h4_tag.append(soup.new_string(text))
        body.append(h4_tag)
        return h4_tag

    def output_p(self, soup, body, text):
        p_tag = soup.new_tag('p')
        p_tag.append(soup.new_string(text))
        body.append(p_tag)




class InterpreterBase:

    def remove_everything_before(self, tag, including=False):
        for sibling in tag.find_previous_siblings():
            sibling.decompose()
        if including: tag.decompose()

    def find_body_ancestor(self, tag):
        if not tag.parent:
            return None, None
        elif tag.parent.name == 'body':
            return tag, tag.parent
        else:
            return self.find_body_ancestor(tag.parent)

    def remove_everything_before_body_level(self, tag, including=False):
        # Find the ancestor element that has body as a parent
        body_parent, body = self.find_body_ancestor(tag)
        if body_parent:
            self.remove_everything_before(body_parent, including)
            return body

    def replace_invalid_tag(self, doc, tag):
        for match in doc.findAll(tag):
            match.unwrap()



from itertools import takewhile
from bs4 import BeautifulSoup, NavigableString

import bikesanity.io_utils.log_handler as log_handler
from bikesanity.entities.content_blocks import TextBlock, Heading, Image, Map
from bikesanity.entities.page import Page
from bikesanity.services.retrievers import BaseRetriever

from bikesanity.services.map_extractor import MapExtractor, LineExtractor
from .interpreter_base import InterpreterBase


class PageInterpreter(InterpreterBase):
    def __init__(self, retriever: BaseRetriever, progress_callback=None):
        self.retriever = retriever
        self.progress_callback = progress_callback

    def update_progress(self):
        if self.progress_callback: self.progress_callback()


    def get_title(self, title):
        title_elem = title.find(name='h1')
        return title_elem.text.strip() if title_elem else ''

    def get_distance_statement(self, title):
        distance_elem = title.find(name='h3')
        if not distance_elem or distance_elem.parent.name != 'body': return ''
        distance_statement = distance_elem.text.strip()
        return distance_statement

    def get_main_content(self, title):
        content_elem = title.find(name='div')
        return content_elem

    def is_picture_or_map(self, div_elem):
        table_elem = div_elem.find('table')
        return table_elem and 'id' in table_elem.attrs and 'pic' in table_elem.attrs['id']

    def is_map(self, div_elem):
        return any(div_elem.findAll("div", {"id" : lambda l: l and l.startswith('div_map')}))

    def remove_children(self, children):
        for child in children:
            if isinstance(child, NavigableString):
                child.extract()
            else:
                child.decompose()

    def get_text_before_next_element(self, content):
        for dodgy_center in content.find_all(name="center", recursive=False):
            dodgy_center.unwrap()

        div_elem = content.find(name='div')

        if div_elem == next(content.children, None):
            return None, div_elem

        text_content = []
        to_remove = []
        for child in takewhile(lambda x: x != div_elem, content.children):
            # Convert any <br> into newlines
            if hasattr(child, 'find_all'):
                for br in child.find_all('br'):
                    br.replace_with('\n' + br.text)

            if hasattr(child, 'get_text'):
                text_content.append( child.get_text().strip())
                to_remove.append(child)
            else:
                text_content.append(str(child).strip())
                to_remove.append(child)

        self.remove_children(to_remove)
        return text_content, div_elem


    def process_pic(self, div_elem):
        pic_elem = div_elem.find(name='center')
        if not pic_elem: return None

        # Get image
        img_elem = pic_elem.find(name='img')
        if not img_elem: return None

        img_path = img_elem.attrs['src']
        img_path_fullsize = img_path.replace('/small/', '/large/')

        # Download the pics
        pic_small = self.retriever.retrieve_image_small(img_path)
        pic_fullsize = self.retriever.retrieve_image_large(img_path_fullsize)

        # Get caption
        caption_elem = pic_elem.find(name='b')
        caption = caption_elem.text if caption_elem else ''

        return Image(img_path, img_path_fullsize, caption, pic_small, pic_fullsize)

    def _get_map_id_from_path(self, map_path):
        for token in map_path.split('?'):
            if token.startswith('line_id='):
                return token[8:]


    def process_map(self, div_elem):
        map_elem = div_elem.find(name='center')
        if not map_elem: return None

        # Retrieve the map resource path
        gpx_download_elem = div_elem.find('a', {'title' : lambda l: l and l.startswith('Download GPX')})
        if gpx_download_elem:
            gpx_path = gpx_download_elem.attrs['href']
            gpx_url = self.retriever.base_url + gpx_path
            # Get the map ID from the path
            map_id = self._get_map_id_from_path(gpx_path)
        else:
            gpx_url = None
            map_id = None

        # Get caption
        caption_elem = map_elem.find(name='b')
        caption = caption_elem.text if caption_elem else ''

        # Now parse out the data directly from the HTML representation of the map
        gpx_data, json_data = None, None
        try:
            map_extractor = MapExtractor(map_elem)
            map_data = next(map_extractor.iterate_map_data(), None)
            if map_data:
                line_extractor = LineExtractor(map_data)
                gpx_data, json_data = line_extractor.generate_map_binary_data()
        except Exception as exc:
            log_handler.log.exception('Error extracting map', exc)

        log_handler.log.info('Successfully extracted map {0}'.format(map_id))
        return Map(map_id, caption, gpx_data=gpx_data, json_data=json_data, url=gpx_url)


    def process_next_bloc(self, main_content, page: Page):
        text_content, elem = self.get_text_before_next_element(main_content)
        if text_content: page.add_content(TextBlock(text_content))

        if not elem: return False

        if self.is_picture_or_map(elem):
            if self.is_map(elem):
                map_block = self.process_map(elem)
                if map_block: page.add_content(map_block)
            else:
                image_block = self.process_pic(elem)
                if image_block: page.add_content(image_block)

        self.remove_everything_before(elem, including=True)
        self.update_progress()
        return True


    def process_title(self, title, page: Page, include_metadata=True, include_title=False):

        if include_metadata:
            page.title = self.get_title(title)
            page.parse_distance_statement(self.get_distance_statement(title))

        if include_title:
            title_text = self.get_title(title)
            if title_text: page.add_content(Heading(title_text))

        main_content = self.get_main_content(title)
        while(main_content):
            if not self.process_next_bloc(main_content, page): break

    def process_titles(self, doc, page: Page, include_metadata=True):
        titles = doc.find_all(name="big")
        for title in titles:
            parent = title.parent
            if parent and parent.name == 'body':
                self.remove_everything_before(title)
                self.process_title(parent, page, include_metadata=include_metadata)

    def process_titles_single_page(self, doc, page: Page):
        titles = doc.find_all(name="hr")
        for title in titles:
            parent = title.parent
            if parent and parent.name == 'td':
                body = self.remove_everything_before_body_level(title, including=True)
                self.process_title(body, page, include_metadata=False, include_title=True)

    def parse_page(self, page: Page, single=False):
        log_handler.log.info('Processing page {0} for journal {1}'.format(page.original_id, page.journal_id))

        doc = BeautifulSoup(page.original_html, features="lxml")
        titles_processor = self.process_titles_single_page if single else self.process_titles
        titles_processor(doc, page)

        # If additional pages exist, process these as well
        for additional_html in page.additional_html_pages:
            log_handler.log.info('Processing additional page for journal {0}'.format(page.journal_id))
            doc = BeautifulSoup(additional_html, features="lxml")
            self.process_titles(doc, page, include_metadata=False)

        return page

    def get_additional_part_id(self, url):
        for term in url.split('&'):
            if term.startswith('part='):
                return term[5:]

    def find_additional_pages(self, page_html):
        doc = BeautifulSoup(page_html, features="lxml")

        titles = doc.find_all(name='big')
        for title in titles:
            page_link_elem = title.find(name='a')
            if page_link_elem and '>>>' in title.get_text():
                url = self.retriever.base_url + page_link_elem.attrs['href']
                yield url, self.get_additional_part_id(url)


    def retrieve_page(self, journal_id, original_id, url):
        log_handler.log.info('Retrieving page {0} for journal {1}'.format(original_id, journal_id), extra={'journal_id': journal_id})
        self.update_progress()

        # Download the page HTML
        page_html = self.retriever.retrieve_page(url, original_id, error_message="Error code in retrieving journal page html: {0}")
        self.update_progress()

        page = Page(journal_id, original_id, page_html)

        # Download any additional other pages
        for url, part_id in self.find_additional_pages(page_html):
            log_handler.log.info('Additional page for {0} detected: {1}'.format(original_id, url), extra={'journal_id': journal_id})
            self.update_progress()

            # Lack of a parsed part number means this has already been postprocessed (and then just use the whole path)
            additional_filename = '{0}_{1}'.format(original_id, part_id) if part_id else url.replace('.html', '')

            additional_html = self.retriever.retrieve_page(url, additional_filename, error_message="Error code in retrieving additional page html: {0}")
            page.add_additional_html(additional_html)
            self.update_progress()

        return page



from itertools import takewhile
import unicodedata

from bs4 import BeautifulSoup, NavigableString

import bikesanity.io_utils.log_handler as log_handler
from bikesanity.services.retrievers import BaseRetriever
from bikesanity.entities.content_blocks import Image
from bikesanity.entities.journal import Journal
from .interpreter_base import InterpreterBase


class JournalContent(InterpreterBase):

    MAX_PAGE_DOWNLOAD_ATTEMPTS = 2


    def __init__(self, retriever: BaseRetriever):
        self.retriever = retriever

    def find_title(self, doc):
        title = doc.find(name="h1")
        self.remove_everything_before(title)
        return title

    def get_subtitle(self, doc):
        subtitle_elem = doc.find('h3')
        if subtitle_elem:
            self.remove_everything_before(subtitle_elem)
            return subtitle_elem.get_text()
        return None

    def get_distance_statement(self, doc):
        subtitle_elem = doc.find('h3')
        if subtitle_elem and isinstance(subtitle_elem.next_sibling, NavigableString):
            return subtitle_elem.next_sibling.replace('\n', '').replace('\t', '').strip()

    def get_author_statement(self, doc):
        title = doc.title.text
        if 'by ' in title:
            return title[title.rfind('by ')+3:]
        else:
            author_elem = doc.find('a', {"href" : lambda l: l and '&user' in l})
            return author_elem.get_text() if author_elem else None

    def get_locales(self, doc):
        locales_elem = doc.find('b', string="Locales:") or doc.find('b', string="Locale:")
        locales = []
        if locales_elem:
            for sibling in locales_elem.find_next_siblings():
                if sibling.name != 'a': break
                locale_name = sibling.get_text()
                locale_name = unicodedata.normalize("NFKD", locale_name)
                locales.append(locale_name)
        return locales

    def get_cover_pic(self, doc):
        pic_containers = doc.findAll("table", {"bgcolor" : lambda l: l and l == 'gainsboro'})
        for pic_container in pic_containers:
            img_elem = pic_container.find('img', {'src': lambda l: l and 'pics/' in l})
            if img_elem:
                img_path = img_elem.attrs['src']
                cover_pic_data = self.retriever.retrieve_image_large(img_path)

                caption_elem = pic_container.find('b')
                caption = caption_elem.get_text() if caption_elem else ''

                return Image(img_path, img_path, caption, cover_pic_data, cover_pic_data)


    def get_contents_table(self, doc, journal: Journal):
        content_list_elem = doc.find('dl')

        # Handle empty documents with no ToC gracefully
        if not content_list_elem:
            log_handler.log.warn('Empty table of contents - continuing')
            return

        # Replace all the <dd> tags which are completely invalid, well done Neil!
        self.replace_invalid_tag(content_list_elem, 'dd')

        for child in takewhile(lambda x: x, content_list_elem.children):
            # Ignore basic strings
            if isinstance(child, NavigableString): continue

            # Find whether the ToC entry has a link, or if it's just a title
            toc_link = child if 'a' == child.name else child.find('a')

            # Any element that doesn't have an id attribute can be skipped
            if (not 'id' in child.attrs) and (not toc_link or not 'id' in toc_link.attrs): continue

            # Get the URL and ID
            toc_link_url = self.retriever.base_url + toc_link.attrs['href'] if toc_link else None
            toc_id = toc_link.attrs['id'] if toc_link else child.attrs['id']

            # Get the title text, and any other text following it
            toc_title = child.get_text()
            toc_title = toc_title + child.next_sibling if isinstance(child.next_sibling, NavigableString) else toc_title
            toc_title = toc_title.strip()

            journal.add_toc_entry(toc_id, toc_title, toc_link_url)

    def journal_from_html(self, journal_id, html) -> Journal:
        journal = Journal(journal_id, html)
        doc = BeautifulSoup(html, features="lxml")

        # Title
        title = self.find_title(doc)
        journal.journal_title = title.get_text()

        #  Subtitle
        journal.journal_subtitle = self.get_subtitle(doc)

        # Distance statement
        journal.distance_statement = self.get_distance_statement(doc)

        # Author user
        journal.journal_author = self.get_author_statement(doc)

        # Locales
        journal.locales = self.get_locales(doc)

        # Cover pic
        journal.add_cover_image(self.get_cover_pic(doc))

        # Parse the ToC
        self.get_contents_table(doc, journal)

        return journal


    def _get_journal_id_from_url(self, url):
        tokens = url.split('&')
        for token in tokens:
            if token.startswith('doc_id='):
                return token[7:]

    def retrieve_journal(self, url, journal_id) -> Journal:
        # Resolve the full URL against the server and retrieve the journal index
        retrieved_html, resolved_url = self.retriever.retrieve_index(url, error_message="Error getting journal: {0}")

        # Determine the journal ID from the resolved URL
        if not journal_id: journal_id = self._get_journal_id_from_url(resolved_url)

        return self.journal_from_html(journal_id, retrieved_html)

    JS = [
        'leaflet.js',
        'Polyline.encoded.js',
        'plotly-latest.min.js',
        'Leaflet.fullscreen.min.js',
        'json2.js'
    ]
    CSS = [
        'leaflet.css',
        'leaflet.fullscreen.css'
    ]

    def retrieve_js_css_resources(self, journal) -> Journal:
        for js in self.JS:
            journal.js[js] = self.retriever.retrieve_js(js)
        for css in self.CSS:
            journal.css[css] = self.retriever.retrieve_css(css)
        return journal




import datetime
import random
import time

from bikesanity.io_utils import log_handler as log_handler
from .base_session import BaseSession


class ThrottledSession(BaseSession):

    STANDARD_REQUEST_RATE_LIMITER = 25
    STANDARD_REQUEST_RATE_LIMITER_MIN = 25
    STANDARD_REQUEST_RATE_LIMITER_MAX = 30
    FAILED_REQUEST_RATE_LIMITER = 5

    def __init__(self):
        super().__init__()
        self.last_request = None
        self.current_rate_limit = self.STANDARD_REQUEST_RATE_LIMITER

    def _stochastic_delay(self):
        return random.randrange(self.STANDARD_REQUEST_RATE_LIMITER_MIN, self.STANDARD_REQUEST_RATE_LIMITER_MAX)

    def _wait_since_last_request(self):
        rate_limit_delay = self._stochastic_delay()

        if not self.last_request:
            self.last_request = datetime.datetime.now()
        while (datetime.datetime.now() - self.last_request).total_seconds() < rate_limit_delay:
            time.sleep(0.2)
        self.last_request = datetime.datetime.now()

    def make_request(self, url):
        super().make_request(url)
        try:
            self._wait_since_last_request()
            return self.session.get(url, headers=self.headers)
        except Exception as exc:
            # Log the exception, delay a while, then raise
            log_handler.log.error("Error connecting when downloading {0}".format(url))
            time.sleep(self.FAILED_REQUEST_RATE_LIMITER)
            raise

    def make_stream_request(self, url):
        super().make_stream_request(url)
        return self.session.get(url, headers=self.headers, stream=True)



import hashlib
import shutil
import os
import mimetypes


class FileHandler:

    HASH_BUFFER_SIZE = 6553600

    def __init__(self):
        pass

    def output_binary_to_file(self, filename, data):
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        data.seek(0)
        with open(filename, 'wb') as handle:
            shutil.copyfileobj(data, handle)

    def output_bytes_to_file(self, filename, bytes):
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, 'wb') as handle:
            handle.write(bytes)

    def output_text_to_file(self, filename, text):
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, 'w') as handle:
            handle.write(text)

    def remove_directory(self, path):
        try:
            shutil.rmtree(path)
        except:
            pass

    def file_exists(self, filename):
        return os.path.isfile(filename)

    def get_file_content(self, filename):
        if not self.file_exists(filename):
            raise RuntimeError('File did not exist: {0}'.format(filename))
        with open(filename, 'r', encoding="utf8") as handle:
            return handle.read()

    def get_binary_content(self, filename):
        if not self.file_exists(filename):
            raise RuntimeError('File did not exist: {0}'.format(filename))
        with open(filename, 'rb') as handle:
            return handle.read()

    def get_binary(self, filename):
        if not self.file_exists(filename):
            raise RuntimeError('File did not exist: {0}'.format(filename))
        return open(filename, 'rb')

    def file_size(self, filename):
        return os.path.getsize(filename)

    def calculate_sha1_hash(self, filename):
        sha1 = hashlib.sha1()

        with open(filename, 'rb') as f:
            while True:
                data = f.read(self.HASH_BUFFER_SIZE)
                if not data:
                    break
                sha1.update(data)

        return sha1.hexdigest()

    def extension_from_type(self, mime_type):
        extension = mimetypes.guess_extension(mime_type)
        if extension and extension[0] == '.': extension = extension[1:]
        return extension





import uuid


class ContentBlock:
    def clear_resources(self):
        pass

    def _sanitize_path(self, path):
        path = path.split('?')[0]
        if path.startswith('/'): path = path[1:]
        return path


class TextBlock(ContentBlock):
    def __init__(self, content_text):
        self.content_text = content_text


class Heading(ContentBlock):
    def __init__(self, content_text):
        self.content_text = content_text


class Image(ContentBlock):
    def __init__(self, original_path_small, original_path_fullsize, caption, image_small, image_fullsize):
        self.image_id = str(uuid.uuid4())

        self.original_path_small = self._sanitize_path(original_path_small)
        self.original_path_fullsize = self._sanitize_path(original_path_fullsize)
        self.extension = self._find_extension(original_path_small)

        self.caption = caption

        self.image_small = image_small
        self.image_fullsize = image_fullsize

    def _find_extension(self, path):
        if '.' in path:
            path = path[path.rfind('.')+1:]
            return path.split('?')[0].lower()

    def clear_resources(self):
        if self.image_small: self.image_small.close()
        if self.image_fullsize: self.image_fullsize.close()
        self.image_small, self.image_fullsize = None, None

    def get_small_resource_path(self):
        return '{0}.small.{1}'.format(self.image_id, self.extension)

    def get_fullsize_resource_path(self):
        return '{0}.{1}'.format(self.image_id, self.extension)


class Map(ContentBlock):
    def __init__(self, original_id, caption, gpx_data=None, json_data=None, url=None):
        self.map_id = str(uuid.uuid4())
        self.original_id = original_id
        self.caption = caption
        self.gpx_data = gpx_data
        self.json_data = json_data
        self.original_url = url

    def set_gpx_data(self, gpx_data):
        self.gpx_data = gpx_data

    def set_map_data(self, gpx_data, json_data):
        self.gpx_data = gpx_data
        self.json_data = json_data

    def clear_resources(self):
        if self.gpx_data: self.gpx_data.close()
        if self.json_data: self.json_data.close()
        self.gpx_data = None
        self.json_data = None

    def get_resource_path(self):
        return '{0}.gpx'.format(self.map_id)



import os
from pathlib import Path
from bikesanity.processing.publish_journal import PublishJournal, PublicationFormats
from bikesanity.io_utils.log_handler import init_logging, log


BASE_DIRECTORY = 'CycleSanityJournals'
base_path = os.path.join(Path.home(), BASE_DIRECTORY)

input_path = base_path
output_path = base_path

journal_id = '14924'

init_logging()

journal_publisher = PublishJournal(input_path, output_path, journal_id)
journal_publisher.publish_journal_id(PublicationFormats.PDF)



from abc import ABC, abstractmethod
import logging
import requests
import random
import math

from .custom_retry import CustomRetry
from .timeout_http_adaptor import TimeoutHTTPAdapter
from .resources import unserialize_resource_stream


log = logging.getLogger(__name__)


class BaseSession(ABC):

    BACKOFF_FACTOR = 10
    TOTAL_RETRIES = 2
    REQUEST_TIMEOUT = 20

    SERIALIZED_RESOURCES = 'serialized.pickle'

    def __init__(self):
        self.session = None
        self.user_agent = self.generate_random_user_agent()
        self.headers = unserialize_resource_stream([self.SERIALIZED_RESOURCES])['HEADERS']

    def generate_random_user_agent(self):
        return random.choice(unserialize_resource_stream([self.SERIALIZED_RESOURCES])['USER_AGENTS'])

    def connect_session(self):
        session = requests.session()

        retry_strategy = CustomRetry(
            total=self.TOTAL_RETRIES,
            status_forcelist=[429, 500, 502, 503, 504, 403],
            method_whitelist=["HEAD", "GET", "OPTIONS"],
            backoff_factor=self.BACKOFF_FACTOR
        )

        # Mount it for both http and https usage
        adapter = TimeoutHTTPAdapter(timeout=self.REQUEST_TIMEOUT)
        adapter.max_retries = retry_strategy
        session.mount("https://", adapter)
        session.mount("http://", adapter)

        self.session = session


    def get_session(self):
        return self.session

    def handle_cookies(self):
        browser_cookie = self.session.cookies.get('browser')
        browserx_cookie = self.session.cookies.get('browserx')

        if not browserx_cookie:
            browserx_cookie = str(math.floor(random.random() * 10000000))
            self.session.cookies.set('browserx', browserx_cookie, domain='.crazyguyonabike.com')

        if not browser_cookie or '.' in browser_cookie: return

        browser_cookie = '{0}.{1}'.format(browser_cookie, browserx_cookie)
        log.warn('Setting cookies: {0}'.format(browser_cookie))
        self.session.cookies.set('browser', browser_cookie, domain='.crazyguyonabike.com')

    @abstractmethod
    def make_request(self, url):
        self.headers["User-Agent"] = self.user_agent
        self.handle_cookies()

    @abstractmethod
    def make_stream_request(self, url):
        self.headers["User-Agent"] = self.user_agent
        self.handle_cookies()



import tempfile
import pickle
import bikesanity
import pkg_resources


RESOURCE_DIRECTORY = 'resources'
resource_package = bikesanity.__name__


def _resource_path(paths):
    paths.insert(0, RESOURCE_DIRECTORY)
    return '/'.join(paths)


def get_resource_string(paths):
    resource_path = _resource_path(paths)
    data = pkg_resources.resource_string(resource_package, resource_path)
    return data.decode()


def get_resource_stream(paths):
    resource_path = _resource_path(paths)
    return pkg_resources.resource_stream(resource_package, resource_path)

def create_temp_from_resource(paths):
    resource_path = _resource_path(paths)
    with pkg_resources.resource_stream(resource_package, resource_path) as stream:
        temp_file = tempfile.NamedTemporaryFile(delete=False)
        temp_file.write(stream.read())
        temp_file.close()
        return temp_file.name


def unserialize_resource_stream(paths):
    with get_resource_stream(paths) as stream:
        return pickle.load(stream)



import logging



LOG_STDOUT_FORMAT = "%(asctime)s [%(levelname)s] %(message)s"


def init_logging():
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s] %(message)s"))

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.addHandler(stream_handler)


log = logging.getLogger()



import re

from bikesanity.entities.content_blocks import Image
from bikesanity.entities.journal import Journal
from bikesanity.entities.page import Page


class HtmlPostProcessor:

    def __init__(self):
        pass

    def postprocess_journal(self, journal: Journal):
        if not journal.original_html: return

        html = journal.original_html.decode()

        # Cover image
        if journal.cover_image:
            html = self.replace_image(html, journal.cover_image)

        # Table of contents links
        for toc in journal.toc:
            if toc.original_id:
                html = self.replace_toc_link(html, toc)

        journal.postprocessed_html = html.encode()

    def postprocess_page(self, page: Page):
        # Postprocess the page itself
        page_html = page.original_html.decode()
        page.postprocessed_html = self.postprocess_page_html(page, page_html)

        # Postprocess additional page HTML, if any
        page.additional_postprocessed_html = []
        for additional_page in page.additional_html_pages:
            additional_html = additional_page.decode()
            page.additional_postprocessed_html.append(self.postprocess_page_html(page, additional_html))


    def postprocess_page_html(self, page: Page, page_html):
        # Fix JS and CSS links
        page_html = self.fix_js_css_links(page_html)

        # Fix multipage links
        page_html = self.replace_multipage_part_links(page_html, page)

        # Switch maps to use the OSM basemap so they work locally
        page_html = self.switch_maps_to_osm(page_html)

        # Post process images and links
        for content in page.contents:
            if isinstance(content, Image):
                page_html = self.replace_image(page_html, content)

        return page_html.encode()


    def replace_toc_link(self, html, toc) -> str:
        pattern = '<a href=\\"/doc/page/.*?ID=\\"{0}\\">'.format(toc.original_id)
        replacement = '<a href="{0}.html" ID="{1}">'.format(toc.original_id, toc.original_id)
        return re.sub(pattern, replacement, html)

    def replace_image(self, html, image: Image):
        image_path = re.escape(image.original_path_small)
        pattern = '<a href=\\".*\\">\\s*<img src=\\"/{0}.*?\\"'.format(image_path)
        replacement = '<a href="{0}"><img src="{1}"'.format(image.original_path_fullsize, image.original_path_small)
        return re.sub(pattern, replacement, html)

    def switch_maps_to_osm(self, page_html):
        return re.sub('\"basemap\":\"stadia\"', '"basemap":"osm"', page_html)

    def fix_js_css_links(self, page_html):
        page_html = re.sub('<script src=\\"/javascript', '<script src="javascript', page_html)
        page_html = re.sub("<script src='/javascript", "<script src='javascript", page_html) # Yes, really. NG is an idiot.
        page_html = re.sub('<script type=\\"text/javascript\\" src=\\"/javascript', '<script type="text/javascript" src="javascript', page_html)
        page_html = re.sub('<link rel=\\"stylesheet\\" href=\\"/css/', '<link rel="stylesheet" href="css/', page_html)
        page_html = re.sub("<link href='/css", "<link href='css", page_html)
        return page_html

    def replace_multipage_part_links(self, page_html, page: Page):
        page_html = re.sub(r'(>>> <A HREF=\")(.*?)(\d+)(\">)', r'\g<1>{0}_\g<3>.html\g<4>'.format(page.original_id), page_html)
        page_html = re.sub(r'(<<< <A HREF=\")(.*?)(\">)', r'\g<1>{0}.html\g<3>'.format(page.original_id), page_html)
        page_html = re.sub(r'(<<< <A HREF=\")(.*?)(\d+)(\">)', r'\g<1>{0}_\g<3>.html\g<4>'.format(page.original_id), page_html)
        return page_html



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bikesanity",
    version="1.1.5",
    author="John Henry",
    author_email="john.henry.split.my.heart@protonmail.com",
    description="BikeSanity cycle touring journal extractor and formatter",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JohnHenrySplitMyHeart/bikesanity",
    download_url = 'https://github.com/JohnHenrySplitMyHeart/bikesanity/archive/v_1.1.5.tar.gz',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'beautifulsoup4>=4.8.2',
        'requests>=2.22.0',
        'lxml>=4.5.2',
        'html5lib>=1.1',
        'gpxpy>=1.4.2',
        'Click>=7.0',
        'jsonpickle>=1.5.1',
        'fpdf>=1.7.2',
        'prettierfier>=1.0.3'
    ],
    entry_points={
        'console_scripts': ['bikesanity-run=bikesanity.run:run'],
    },
    include_package_data=True
)



import pickle
from io import BytesIO

from bikesanity.entities.journal import Journal


class Serializer:

    def serialize_journal(self, journal: Journal):
        buffer = BytesIO()
        pickle.dump(journal, buffer)
        buffer.seek(0)
        return buffer

    def serialize_journal_to_disk(self, filename, journal: Journal):
        with open(filename, 'wb') as handle:
            pickle.dump(journal, handle)

    def serialize_to_disk(self, filename, obj):
        with open(filename, 'wb') as handle:
            pickle.dump(obj, handle)

    def serialize_to_byteio(self, obj):
        buffer = BytesIO()
        pickle.dump(obj, buffer)
        buffer.seek(0)
        return buffer

    def unserialize_from_data(self, data):
        try:
            return pickle.load(data)
        except Exception as exc:
            return None
        finally:
            if data: data.close()



from enum import Enum, auto


class SectionType(Enum):
    TEXT = auto()
    PIC = auto()


class JsonSection:
    def __init__(self, section_type: SectionType):
        self.sectionType = section_type
        self.text = None
        self.picId = None
        self.picFilename = None
        self.picUrl = None
        self.picCaption = None

    @staticmethod
    def text_section(text):
        section = JsonSection(SectionType.TEXT)
        section.text = text
        return section

    @staticmethod
    def pic_section(id, filename, url, caption):
        section = JsonSection(SectionType.PIC)
        section.picId = id
        section.picFilename = filename
        section.picUrl = url
        section.picCaption = caption
        return section


class JsonDistance:
    def __init__(self, miles=None, km=None, totalMilesSoFar=None, totalKmSoFar=None):
        self.miles = miles
        self.km = km
        self.totalMilesSoFar = totalMilesSoFar
        self.totalKmSoFar = totalKmSoFar


class JsonPageContents:
    def __init__(self):
        self.isHeader = None
        self.title = None
        self.date = None
        self.distance = None
        self.sections = []

    def set_header(self, header):
        self.isHeader = True
        self.title = header

    def set_page(self, title, date):
        self.isHeader = False
        self.title = title
        self.date = date

    def set_distance(self, distance: JsonDistance):
        self.distance = distance

    def add_section(self, section: JsonSection):
        self.sections.append(section)


class JsonPage:
    def __init__(self, id, url):
        self.id = id
        self.url = url
        self.contents = JsonPageContents()


class JsonJournal:

    def __init__(self, doc_id, title, subtitle):
        self.docId = doc_id
        self.title = title
        self.subtitle = subtitle

        self.distanceInMiles = None
        self.distanceInKilometers = None
        self.primaryDistanceUomIsMiles = True
        self.firstDate = None
        self.lastDate = None
        self.lastUpdate = None

        self.authors = []
        self.pages = []

    def add_author(self, author):
        self.authors.append(author)

    def add_page(self, page: JsonPage):
        self.pages.append(page)



import logging
from .content_blocks import ContentBlock, Image, Map

log = logging.getLogger(__name__)


class Page:

    def __init__(self, journal_id, original_id, original_html):
        self.journal_id = journal_id
        self.original_id = original_id

        self.original_html = original_html
        self.postprocessed_html = None

        self.additional_html_pages = []
        self.additional_postprocessed_html = []

        self.title = None
        self.date_statement = None
        self.page_distance = None
        self.total_distance = None

        self.contents = []
        self.maps = []

    def add_additional_html(self, additional_html):
        self.additional_html_pages.append(additional_html)

    def add_content(self, content_block: ContentBlock):
        self.contents.append(content_block)

    def parse_distance_statement(self, distance_statement):
        if not distance_statement: return

        if 'Total so far: ' in distance_statement:
            self.total_distance = distance_statement[distance_statement.find('Total so far: ')+14:].strip()

            self.page_distance = distance_statement[:distance_statement.find(' - ')]
            self.page_distance = self.page_distance[self.page_distance.rfind(', ')+2:].strip()
            self.date_statement = distance_statement[:distance_statement.find(self.page_distance)][:-3].strip()
        else:
            self.date_statement = distance_statement

    def _persist_resources(self, persist_image, persist_map):
        self.maps = []

        # Iterate over content and upload
        for content in self.contents:
            if isinstance(content, Image):
                persist_image(content)
            elif isinstance(content, Map):
                persist_map(content)

    def save_resources(self, local_handler):
        self._persist_resources(local_handler.save_image_resource, local_handler.save_map_resource)

    def save_originals(self, local_handler):
        # Save the page HTML, including any additional sections
        local_handler.save_html_original(self.original_id + '.html', self.postprocessed_html if self.postprocessed_html else self.original_html)
        part = 2

        additional_html_pages = self.additional_postprocessed_html if self.additional_postprocessed_html else self.additional_html_pages
        for additional_html in additional_html_pages:
            filename = '{0}_{1}.html'.format(self.original_id, part)
            local_handler.save_html_original(filename, additional_html)
            part += 1

        # Save the associated image and map originals
        self._persist_resources(local_handler.save_image_original, local_handler.save_map_original)

    def clear_resources(self):
        for content in self.contents:
            content.clear_resources()
